@@ -15,9 +15,11 @@
  */
 package net.sf.cglib.proxy;
 
+import java.io.ByteArrayInputStream;
 import java.io.ByteArrayOutputStream;
 import java.io.IOException;
 import java.io.InputStream;
+import java.io.Serializable;
 import java.lang.ref.PhantomReference;
 import java.lang.ref.ReferenceQueue;
 import java.lang.reflect.Field;
@@ -25,16 +27,22 @@ import java.lang.reflect.Method;
 import java.lang.reflect.Modifier;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.HashMap;
 import java.util.List;
-
+import java.util.Map;
 import junit.framework.Test;
 import junit.framework.TestCase;
 import junit.framework.TestSuite;
 import net.sf.cglib.CodeGenTestCase;
 import net.sf.cglib.core.AbstractClassGenerator;
 import net.sf.cglib.core.DefaultNamingPolicy;
+import net.sf.cglib.core.NamingPolicy;
+import net.sf.cglib.core.Predicate;
 import net.sf.cglib.core.ReflectUtils;
 import net.sf.cglib.reflect.FastClass;
+import org.objectweb.asm.ClassWriter;
+import org.objectweb.asm.MethodVisitor;
+import org.objectweb.asm.Opcodes;
 
 /**
  *@author     Juozas Baliuka <a href="mailto:baliuka@mwm.lt">
@@ -159,13 +167,6 @@ public class TestEnhancer extends CodeGenTestCase {
 
         Factory factory = (Factory)proxy;
         assertEquals("((EA)factory.newInstance(factory.getCallbacks())).getName()", "herby", ((EA)factory.newInstance(factory.getCallbacks())).getName());
-
-        Enhancer e = new Enhancer();
-        e.setSuperclass(EA.class);
-        e.setCallbackType(MethodInterceptor.class);
-        Factory factory1 = e.createFactory();
-        assertEquals("((EA)e.createFactory().newInstance(factory.getCallbacks())).getName()",
-                "herby", ((EA)factory1.newInstance(new DelegateInterceptor(save))).getName());
     }
 
     class DelegateInterceptor implements MethodInterceptor {
@@ -268,6 +269,49 @@ public class TestEnhancer extends CodeGenTestCase {
         
     }
 
+    public void testProxyClassReuseAcrossGC() throws InterruptedException {
+        String proxyClassName = null;
+        for (int i = 0; i < 50; i++) {
+            Enhancer enhancer = new Enhancer();
+            enhancer.setSuperclass(Source.class);
+            enhancer.setCallbackFilter(new CallbackFilter() {
+                public int accept(Method method) {
+                    return 0;
+                }
+
+                @Override
+                public boolean equals(Object obj) {
+                    return true;
+                }
+
+                @Override
+                public int hashCode() {
+                    return 0;
+                }
+            });
+            enhancer.setInterfaces(new Class[]{Serializable.class});
+            enhancer.setCallback(new InvocationHandler() {
+                public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
+                    if (method.getDeclaringClass() != Object.class
+                            && method.getReturnType() == String.class) {
+                        return null;
+                    } else {
+                        throw new RuntimeException("Do not know what to do.");
+                    }
+                }
+            });
+            Source proxy = (Source) enhancer.create();
+            String actualProxyClassName = proxy.getClass().getName();
+            if (proxyClassName == null) {
+                proxyClassName = actualProxyClassName;
+            } else {
+                assertEquals("GC iteration " + i + ", proxy class should survive GC and be reused even across GC",
+                        proxyClassName, actualProxyClassName);
+            }
+            System.gc();
+        }
+    }
+
     /**
      * Verifies that the cache in {@link AbstractClassGenerator} SOURCE doesn't
      * leak class definitions of classloaders that are no longer used.
@@ -636,6 +680,51 @@ public class TestEnhancer extends CodeGenTestCase {
       dummy.toString();
       assertTrue(ran[0]);
     }
+    
+    public void testBadNamingPolicyStillReservesNames() throws Throwable {
+      Enhancer e = new Enhancer();
+      e.setUseCache(false);
+      e.setCallback(NoOp.INSTANCE);
+      e.setClassLoader(new ClassLoader(this.getClass().getClassLoader()){});
+      e.setNamingPolicy(new NamingPolicy() {      
+        public String getClassName(String prefix, String source, Object key, Predicate names) {
+          return "net.sf.cglib.empty.Object$$ByDerby$$123";
+        }
+      });
+      Class proxied = e.create().getClass();
+      final String name = proxied.getCanonicalName();
+      final boolean[] ran = new boolean[1];
+      e.setNamingPolicy(new NamingPolicy() {
+        public String getClassName(String prefix, String source, Object key, Predicate names) {
+          ran[0] = true;
+          assertTrue(names.evaluate(name));
+          return name + "45"; 
+        }
+      });
+      Class proxied2 = e.create().getClass();
+      assertTrue(ran[0]);
+      assertEquals(name + "45", proxied2.getCanonicalName());
+    }
+
+    /**
+     * In theory, every sane implementation of {@link NamingPolicy} should check if the class name is occupied,
+     * however, in practice there are implementations in the wild that just return whatever they feel is good.
+     *
+     * @throws Throwable if something wrong happens
+     */
+    public void testNamingPolicyThatReturnsConstantNames() throws Throwable {
+      Enhancer e = new Enhancer();
+      final String desiredClassName = "net.sf.cglib.empty.Object$$42";
+      e.setCallback(NoOp.INSTANCE);
+      e.setClassLoader(new ClassLoader(this.getClass().getClassLoader()){});
+      e.setNamingPolicy(new NamingPolicy() {
+        public String getClassName(String prefix, String source, Object key, Predicate names) {
+          return desiredClassName;
+        }
+      });
+      Class proxied = e.create().getClass();
+      assertEquals("Class name should match the one returned by NamingPolicy", desiredClassName, proxied.getName());
+    }
 
     public static Object enhance(Class cls, Class interfaces[], Callback callback, ClassLoader loader) {
         Enhancer e = new Enhancer();
@@ -1172,9 +1261,163 @@ public class TestEnhancer extends CodeGenTestCase {
         intf.aMethod(null);
         assertEquals(Arrays.asList(Concrete.class), paramTypes);
     }
-    
-    
-    
+
+    public void testBridgeParameterCheckcast() throws Exception {
+
+    // If the compiler used for Z omits the bridge method, and X is compiled with javac,
+    // javac will generate an invokespecial bridge in X.
+
+    // public interface I<A, B> {
+    //   public A f(B b);
+    // }
+    // public abstract class Z<U extends Number> implements I<U, Long> {
+    //   public U f(Long id) {
+    //     return null;
+    //   }
+    // }
+    // public class X extends Z<Integer> {}
+
+    final Map<String, byte[]> classes = new HashMap<String, byte[]>();
+
+    {
+      ClassWriter cw = new ClassWriter(0);
+      cw.visit(
+          49,
+          Opcodes.ACC_PUBLIC | Opcodes.ACC_ABSTRACT | Opcodes.ACC_INTERFACE,
+          "I",
+          "<A:Ljava/lang/Object;B:Ljava/lang/Object;>Ljava/lang/Object;",
+          "java/lang/Object",
+          null);
+      {
+        MethodVisitor mv =
+            cw.visitMethod(
+                Opcodes.ACC_PUBLIC | Opcodes.ACC_ABSTRACT,
+                "f",
+                "(Ljava/lang/Object;)Ljava/lang/Object;",
+                "(TB;)TA;",
+                null);
+        mv.visitEnd();
+      }
+      cw.visitEnd();
+      classes.put("I.class", cw.toByteArray());
+    }
+    {
+      ClassWriter cw = new ClassWriter(0);
+      cw.visit(
+          49,
+          Opcodes.ACC_PUBLIC | Opcodes.ACC_SUPER | Opcodes.ACC_ABSTRACT,
+          "Z",
+          "<U:Ljava/lang/Number;>Ljava/lang/Object;LI<TU;Ljava/lang/String;>;",
+          "java/lang/Object",
+          new String[] {"I"});
+      {
+        MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", "()V", null, null);
+        mv.visitCode();
+        mv.visitVarInsn(Opcodes.ALOAD, 0);
+        mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/Object", "<init>", "()V", false);
+        mv.visitInsn(Opcodes.RETURN);
+        mv.visitMaxs(1, 1);
+        mv.visitEnd();
+      }
+      {
+        MethodVisitor mv =
+            cw.visitMethod(
+                Opcodes.ACC_PUBLIC,
+                "f",
+                "(Ljava/lang/String;)Ljava/lang/Number;",
+                "(Ljava/lang/String;)TU;",
+                null);
+        mv.visitCode();
+        mv.visitInsn(Opcodes.ACONST_NULL);
+        mv.visitInsn(Opcodes.ARETURN);
+        mv.visitMaxs(1, 2);
+        mv.visitEnd();
+      }
+      cw.visitEnd();
+      classes.put("Z.class", cw.toByteArray());
+    }
+    {
+      ClassWriter cw = new ClassWriter(0);
+      cw.visit(
+          49, Opcodes.ACC_PUBLIC | Opcodes.ACC_SUPER, "X", "LZ<Ljava/lang/Integer;>;", "Z", null);
+      {
+        MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", "()V", null, null);
+        mv.visitCode();
+        mv.visitVarInsn(Opcodes.ALOAD, 0);
+        mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "Z", "<init>", "()V", false);
+        mv.visitInsn(Opcodes.RETURN);
+        mv.visitMaxs(1, 1);
+        mv.visitEnd();
+      }
+      {
+        MethodVisitor mv =
+            cw.visitMethod(
+                Opcodes.ACC_PUBLIC | Opcodes.ACC_BRIDGE | Opcodes.ACC_SYNTHETIC,
+                "f",
+                "(Ljava/lang/Object;)Ljava/lang/Object;",
+                null,
+                null);
+        mv.visitCode();
+        mv.visitVarInsn(Opcodes.ALOAD, 0);
+        mv.visitVarInsn(Opcodes.ALOAD, 1);
+        mv.visitTypeInsn(Opcodes.CHECKCAST, "java/lang/String");
+        mv.visitMethodInsn(
+            Opcodes.INVOKESPECIAL, "Z", "f", "(Ljava/lang/String;)Ljava/lang/Number;", false);
+        mv.visitInsn(Opcodes.ARETURN);
+        mv.visitMaxs(2, 2);
+        mv.visitEnd();
+      }
+
+      cw.visitEnd();
+
+      classes.put("X.class", cw.toByteArray());
+    }
+
+    ClassLoader classLoader =
+        new ClassLoader(getClass().getClassLoader()) {
+          @Override
+          public InputStream getResourceAsStream(String name) {
+            InputStream is = super.getResourceAsStream(name);
+            if (is != null) {
+              return is;
+            }
+            if (classes.containsKey(name)) {
+              return new ByteArrayInputStream(classes.get(name));
+            }
+            return null;
+          }
+
+          public Class findClass(String name) throws ClassNotFoundException {
+            byte[] ba = classes.get(name.replace('.', '/') + ".class");
+            if (ba != null) {
+              return defineClass(name, ba, 0, ba.length);
+            }
+            throw new ClassNotFoundException(name);
+          }
+        };
+
+    List<Class> retTypes = new ArrayList<Class>();
+    List<Class> paramTypes = new ArrayList<Class>();
+    Interceptor interceptor = new Interceptor(retTypes, paramTypes);
+
+    Enhancer e = new Enhancer();
+    e.setClassLoader(classLoader);
+    e.setSuperclass(classLoader.loadClass("X"));
+    e.setCallback(interceptor);
+
+    Object c = e.create();
+
+    for (Method m : c.getClass().getDeclaredMethods()) {
+      if (m.getName().equals("f") && m.getReturnType().equals(Object.class)) {
+        m.invoke(c, new Object[] {null});
+      }
+    }
+
+    // f(Object)Object should bridge to f(Number)String
+    assertEquals(Arrays.asList(Object.class, Number.class), retTypes);
+    assertEquals(Arrays.asList(Object.class, String.class), paramTypes);
+  }
+
     static class ErasedType {}
     static class RetType extends ErasedType {}
     static class Refined extends RetType {}
